unit Parser;

interface

uses
  Windows, Messages, SysUtils, Variants, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, Math;

type
  TParser = class;

  TVariable = record
    name: string;
    index: integer;
  end;

  TRTProc = procedure(X: TParser);
  TRTProcArr = array of TRTProc;

  TOperator = record
    name: string;
    code: TRTProc;
    stacklengthchange: integer;
  end;
  TOpArr = array of TOperator;

  TComparison = record
    name: string;
    CompMode: integer;
    (*
      Gltige Wetrte:
       0: '='
       1: '>='
       2: '>'
      -1: '<='
      -2: '<'
      -3: '<>'
    *)
  end;
  TCompArr = array of TComparison;

  TParser = class
    CurrentLine: string;
    Width, Height: Integer;
    stack: array of array of array of double;
    Variables: array of TVariable;
    Comparisons: TCompArr;
    Operators: TOpArr;
    BinCode: array of TRTProc;
    IP: Integer;
    CompMode: Integer;
    procedure SaveChs(var line: string);
    procedure IgnoreSign(var line: string; ignore: string);
    procedure SetBrackets(var line: string; operators: array of string);
    procedure GetInstructions(code: string; operators: array of string; out _instructions: string);
    procedure ParseComp(code: string; out instructions: string);

    procedure ParseInstr(code: string; out instructions: string);
    procedure VarName(name: string; index: integer); //erzeugt eine Variable mit angegebenem Index bzw aktualisiert den Index

    procedure Assemble(code: string; Operators: TOpArr; Comparisons: TCompArr; out instructions: TRTProcArr); //bislang nur Vergleiche!!
    procedure compile(code: array of string);

    procedure execute(out index: integer);
    procedure Draw(Canvas: TCanvas);

    constructor Create(aWidth, aHeight: Integer; XLo, XHi, YLo, YHi: double); //Operatoren erweitern!!!
  end;

  procedure Add(X: TParser); //die obersten beiden Zahlen addieren, in die zweitoberste schreiben, und die lnge dekrementieren, IP inkrementieren
  procedure Sub(X: TParser); //die obersten beiden Zahlen subtrahieren, in die zweitoberste schreiben, und die lnge dekrementieren, IP inkrementieren
  procedure Mul(X: TParser);
  procedure Divide(X: TParser); //Division durch 0 erzeugt unabhngig vom Vorzeichen infinity, weil Delphi bei echter division eine exception wirft
  procedure Modulo(X: TParser);
  procedure Pow(X: TParser); //berechnet keine Wurzeln aus negativen Zahlen (KEINE, auch nicht die ungeraden)
  procedure Log(X: TParser);
  procedure Sin(X: TParser); //Sinus der obersten Zahl berechnen, und in die oberste Zahl speichern
  procedure Cos(X: TParser);
  procedure Ln(X: TParser);
  procedure Sgn(X: TParser);
  procedure Abs(X: TParser);
  procedure Trunc(X: TParser);
  procedure asin(X: TParser); //in der Anwendung geschrieben arcsin
  procedure acos(X: TParser);

  procedure Chs(X: TParser);
  procedure PushVar(X: TParser); //kopiert bereits im Stack vorhandene Werte oben auf den Stack; IP um 2 erhhen
  procedure PushConst(X: TParser); //legt eine Konstante, dei sich in den nchsten zwei Anweisungen befindet, oben auf den Stack; IP um 3 erhhen

  function Variable(aName: string; aIndex: integer): TVariable;
  function OperatorMake(aName: string; aCode: TRTProc; aStackLengthChange: integer): TOperator;
  function Comparison(aName: string; aCompMode: integer): TComparison;

implementation

const
  alphabet: set of char = ['a'..'z', 'A'..'Z'];
  numbers: set of char = ['0'..'9'];
  opsigns: set of char = ['+', '-', '*', '/', '^', '=', '<', '>'];
  op_chs = #253;

procedure TParser.SaveChs;
type TCharSet = set of char;
var
  i1, i2: integer;
  c: string;
  b: boolean;
  charset: ^TCharSet;

  function GetLeftExpr(index: Integer): string;
  begin
    charset:= nil;
    result:= '';
    (*if index <= 1 then begin
      result:= '(';
      exit;
    end;*)
    repeat
      dec(index);
      if index = 0 then begin
        if result = '' then result:= '(';
        break;
      end;
      if line[index] = ' ' then if result = '' then continue
                                else break; //Leerzeichen schliet den Ausdruck ab
      if line[index] = '(' then begin
        result:= '(';
        break;
      end; //Klammer auf bedeutet, - ist chs

      if charset = nil then begin
        if line[index] in alphabet then charset:= @alphabet;
        if line[index] in numbers then charset:= @numbers;
        if line[index] in opsigns then charset:= @opsigns;
        result:= line[index]; //muss das erste Zeichen sein
      end
      else begin
        if line[index] in charset^ then begin
          result:= line[index] + result; //zhlt nach links
        end
        else if (charset = @numbers) and (line[index] in alphabet) or (charset = @alphabet) and (line[index] in numbers) then begin
          result:= line[index] + result;
        end
        else break;
      end;
    until false;
  end;

begin
  for i1:= 1 to length(line) do begin
    if line[i1] = '-' then begin
      c:= GetLeftExpr(i1);
      b:= false;
      if c = '(' then b:= true;
      for i2:= 0 to high(Operators) do
        if c = Operators[i2].name then b:= true; 
      if b then begin
        line[i1]:= op_chs;
        while line[i1 + 1] = ' ' do begin
          delete(line, i1 + 1, 1);
        end;//Leerzeichen sorgen dafr, dass chs vor Pushvar aufgerufen wird, und dafr dass chs nicht zur Konstante gehrt
      end;
    end;
  end;
end;

procedure TParser.IgnoreSign;
var len, p: integer;
begin
  len:= length(ignore);
  repeat
    p:= pos(ignore, line);
    if p <= 0 then break;
    delete(line, p, len);
    insert(' ', line, p);
  until false;
end;

procedure TParser.SetBrackets;
var
  line2: string;
  op, op2: string;
  currentOp, i1: integer;
  firstOp, oplen: integer;

  procedure DoubleChar(_char: string);//_char MUSS die Lnge 1 haben!
  var
    firstChr: integer;
  begin
    repeat
      firstchr:= pos(_char, line2);
      if firstchr <= 0 then begin
        exit; //keine Klammern mehr da, die noch nicht verdoppelt wurden
      end;
      insert(_char, line, firstchr);//verdoppeln
      delete(line2, firstchr, 1);
      insert(#255 + #255, line2, firstchr);//_char entfernen, und lnge von line2 an line angleichen
    until false;
  end;

begin
  line2:= line;
  DoubleChar('(');
  DoubleChar(')'); //existierende Klammern verdoppeln, damit nicht das passiert: (a + b) * c -> ((a )+( b) * c))
  line:= '(' + line + ')'; //umklammern, damit die ausgeklammerten operatoren keine ')' vor der ersten '(' verursachen knnen

  for currentOp:= 0 to high(operators) do begin
    line2:= line; //line2 wird fr die Suche bentigt
    op:= operators[currentOP];
    op2:= '';
    for i1:= 0 to length(op) - 1 do begin
      op2:= op2 + #255;
    end;
    oplen:= Length(op);

    repeat
      firstOp:= pos(op, line2);
      if firstOp <= 0 then begin
        break; //Schleife verlassen, wenn Operator nicht gefunden
      end;
      delete(line, firstop, oplen); //operator entfernen
      insert(')' + op + '(', line, firstOp); //ausklammern
      delete(line2, firstop, oplen);
      insert(')' + op2 + '(', line2, firstop); //line2 muss gleich lang sein, darf den operator aber nichtmehr enthalten
    until false;
  end;
end;

procedure TParser.GetInstructions;
var
  ops: array of string;

  procedure SetupOps;
  var i1: integer;
  begin
    for i1:= 0 to high(Operators) do begin
      if not ((Operators[i1][1] in alphabet) or (Operators[i1][1] in numbers)) then begin
        SetLength(ops, length(ops) + 1);
        ops[high(ops)]:= operators[i1];
      end;
    end;
  end;

  function SearchRightBracket(line: string): integer;
  var
    i1, count: integer;
  begin
    count:= 1; //erste Klammer wird bersprungen, daher wird bei 1 begonnen
    for i1:= pos('(', line) + 1 to length(line){lenght=high} do begin
      if line[i1] = '(' then inc(count) //eine ) mehr bentigt
      else if line[i1] = ')' then dec(count);
      if count = 0 then break; //genausoviele ) wie ( gefunden ==> erste Klammer geschlossen
    end;
    result:= i1;
  end;

  function GetExpression(var line: string): string;
  var firstspace, firstop, firstop2, i1: integer;
  begin
    repeat
      firstspace:= pos(' ', line);
      if firstspace = 1 then begin
        delete(line, 1, 1);//Leerzeichen am Anfang unerwnscht
      end;
    until firstspace <> 1;

    firstop:= 0;
    for i1:= 0 to high(ops) do begin //ops = Kopie aller Operatoren, die sonderzeichen sind
      firstop2:= pos(ops[i1], line);
      if (firstop = 0) or (firstop2 < firstop) and (firstop2 > 0) then begin
        firstop:= firstop2;
      end;
    end;
    if firstop = 1 then begin
      for i1:= 2 to length(line){length = high} do begin
        if (line[i1] in alphabet) or (line[i1] in numbers) then break;
      end;
      firstop:= i1;
    end; // falls der Ausdruck selbst ein Sonderzeichen ist

    if (firstop > 0) and (firstop < firstspace) then begin
      firstspace:= firstop;
    end;
    firstspace:= firstspace - 1; // der Operator oder whitespace soll nicht mehr dabei sein

    if firstspace > 0 then begin
    result:= copy(line, 1, firstspace);
    delete(line, 1, firstspace); //nchstes GetExpression soll nicht dasselbe zurckgeben
    end
    else begin //wenn kein Space vorhanden ist, ist das ganze der Ausdruck
      result:= copy(line, 1, high(Integer));
      line:= '';
    end;
  end;

  procedure ParseInstructions(line: string; out instructions: string);
  var left, right, i1: integer;
    expr: string;
    dummy: double;
    instr1, instr2, instr3: string;
  label continueWhileLoop;

  begin
    instructions:= '';
    if line = '' then begin
      exit;
    end;
    left:= pos('(', line);
    if left > 0 then begin
      right:= SearchRightBracket(copy(line, left, high(Integer)));// gezhlt ab left!!
    end
    else begin
      right:= 0;
    end;
    if right + left > length(line) + 1 then right:= 0;
    if (left = 0) and (right = 0) then begin
      while length(line) > 0 do begin
        expr:= GetExpression(line);
        if expr = '' then continue; //kann passieren, dass am Ende nur Leerzeichen stehen ==> expr = ''

        for i1:= 0 to high(operators) do begin
          if expr = operators[i1] then begin
            instructions:= instructions + #254 + 'OP' + operators[i1];
            goto ContinueWhileLoop;
          end;
        end;

        //!
        if expr = op_chs then begin
          instructions:= instructions + #254 + 'OP' + op_chs;
          continue;
        end;
        ///Einfache chs anweisung; GetExpression sollte op_chs mitkopieren;

        if expr[1] in alphabet then begin
          instructions:= instructions + #254 + 'PUSHVAR' + expr;
          continue;
        end;

        if expr[1] in numbers then begin
          if not TryStrToFloat(expr, dummy) then begin
            raise Exception.Create(expr + ' is not a valid number');
          end;
          instructions:= instructions + #254 + 'PUSHCONST' + expr;
          continue;
        end;

        //!
        if expr[1] = op_chs then begin
          if expr[2] in alphabet then begin
            instructions:= instructions + #254 + 'PUSHVAR' + copy(expr, 2, high(Integer)) + #254 + 'OP' + op_chs;
            continue;
          end; //Variable mit Vorzeichenwechsel
          if expr[2] in numbers then begin
            if not TryStrToFloat(copy(expr, 2, high(Integer)), dummy) then begin
              raise Exception.Create(copy(expr, 2, high(Integer)) + ' is not a valid number');
            end;
            instructions:= instructions + #254 + 'PUSHCONST-' + copy(expr, 2, high(Integer));
            continue;
          end; //negative Konstante
        end;
        ///!
        ContinueWhileLoop: //einfach nur continue geht nicht, weil auch in einer for schleife. (bei operatoren)
      end;
    end
    else begin
      ParseInstructions(copy(line, 1, left - 1), instr1); //alles bis zur ersten Klammer parsen
      ParseInstructions(copy(line, left + 1, right - 2), instr2); //alles ZWISCHEN erster ( und zugehriger )
      ParseInstructions(copy(line, left + right, high(Integer)), instr3); //alles NACH der zugehrigen )
      instructions:= instr2 + instr1 + instr3; //das in der Klammer muss zuerst kommen
      //(a) + (b) -> PUSHVARa PUSHVARb OP+
      //sin(a) -> PUSHVARa OPsin
    end;
  end;

begin
  SetupOps;
  ParseInstructions(code, _instructions);
end;

procedure TParser.ParseComp;
var mode: string;

  function GetCompMode(var line: string): string;
  var pe, pg, pl, pge, ple, pne, p, oplen: integer;
  begin
    pe:= pos('=', line);
    pg:= pos('>', line);
    pl:= pos('<', line);
    pne:= pos('<>', line);
    ple:= pos('<=', line);
    pge:= pos('>=', line);
    result:= 'MODE';
    p:= 0;
    oplen:= 0;
    if pne > 0 then begin
      result:= result + '<>';
      if pe > 0 then raise Exception.Create('too many operators');
      p:= pne;
      oplen:= 2;
    end
    else if ple > 0 then begin
      result:= result + '<=';
      if pg > 0 then raise Exception.Create('too many operators');
      p:= ple;
      oplen:= 2;
    end
    else if pge > 0 then begin
      result:= result + '>=';
      if pl > 0 then raise Exception.Create('too many operators');
      p:= pge;
      oplen:= 2;
    end
    else if pe > 0 then begin
      result:= result + '=';
      p:= pe;
      oplen:= 1;
    end
    else if pl > 0 then begin
      result:= result + '<';
      p:= pl;
      oplen:= 1;
    end
    else if pg > 0 then begin
      result:= result + '>';
      p:= pg;
      oplen:= 1;
    end;
    delete(line, p, oplen);
    insert('=', line, p); //operator durch = ersetzen, um spter nicht mehr Rcksicht auf dei Art des Vergleichs nehmen zu mssen
  end;

begin
  SaveChs(code);
  IgnoreSign(code, ',');
  mode:= GetCompMode(code);
  SetBrackets(code, ['=']);
  SetBrackets(code, ['+', '-']);
  SetBrackets(code, ['*', '/', 'mod']);
  SetBrackets(code, ['^', 'log']);
  GetInstructions(code, ['=', '<', '>', '+', '-', '*', '/', 'mod', '^', 'log', 'sin', 'cos', 'ln', 'sgn', 'abs', 'trunc', 'arcsin', 'arccos'], instructions);
  instructions:= mode + instructions;
end;

procedure TParser.ParseInstr;
begin
  SaveChs(code);
  IgnoreSign(code, ',');
  SetBrackets(code, ['+', '-']);
  SetBrackets(code, ['*', '/', 'mod']);
  SetBrackets(code, ['^', 'log']);
  GetInstructions(code, ['+', '-', '*', '/', 'mod', '^', 'log', 'sin', 'cos', 'ln', 'sgn', 'abs', 'trunc', 'arcsin', 'arccos'], instructions);
end;

procedure TParser.VarName;
var i1: integer;
begin
  for i1:= 0 to high(Variables) do begin
    if Variables[i1].name = name then begin
      Variables[i1].index:= index;
      exit;
    end;
  end;
  SetLength(Variables, length(Variables) + 1);
  Variables[high(Variables)]:= Variable(name, index);
end;

procedure TParser.Assemble;
  procedure AddInstr(X: TRTProc);
  begin
    SetLength(instructions, length(instructions) + 1);
    instructions[high(instructions)]:= X;
  end;

  procedure OperatorInstr(X: string);
  var i1: integer;
  begin
    for i1:= 0 to high(Operators) do begin
      if X = Operators[i1].name then begin
        if @operators[i1].code <> nil then AddInstr(Operators[i1].code); //fr ignore Zeichen
        break;
      end;
    end;
    if X = op_chs then AddInstr(Chs); //chs kann nicht in den Operatoren stehen, muss deshslb extra geprft werden. 
  end;

  procedure PushVarInstr(X: string);
  var i1, index: integer;
  label BreakForLoop;
  begin
    AddInstr(PushVar);
    for i1:= 0 to high(Variables) do begin
      if X = Variables[i1].name then begin
        index:= Variables[i1].index;
        goto BreakForLoop;
      end;
    end;
    raise Exception.Create('unknown Var identifier: ' + X);
    BreakForLoop:
    AddInstr(TRTProc(index));
  end;

  procedure PushConstInstr(X: string);
  var
    val: double;
    valBits: array[0..1] of integer absolute val;
  begin
    AddInstr(PushConst);
    if not TryStrToFloat(X, val) then begin
      raise Exception.Create('no valid number: ' + X);
    end;
    AddInstr(TRTProc(valBits[0]));
    AddInstr(TRTProc(valBits[1]));
  end;

  function GetInstr(var X: string): string;
  var p: integer;
  begin
    p:= pos(#254, X);
    if p > 0 then begin
      result:= copy(X, 1, p - 1); //ales bis zum ersten #254
      delete(X, 1, p); //alles bis inklusive #254 lschen
    end
    else begin
      result:= copy(X, 1, high(Integer));
      X:= '';
    end; //letzte Anweisung wird nicht durch #254 abgeschlossen
  end;

  //Rechenoperationen compilieren;
  procedure IdentifyInstr(line: string);
  var Instr: string;
  begin
    while length(line) > 0 do begin
      Instr:= GetInstr(line);
      if Instr = '' then continue; //Wrde sonst eine Exception werfen
      if pos('OP', Instr) = 1 then begin
        delete(Instr, 1, 2); // die Zeichen 'OP' werden nicht mehr gebraucht
        OperatorInstr(Instr);
      end
      else if pos('PUSHVAR', Instr) = 1 then begin
        delete(Instr, 1, 7); //die Zeichen 'PUSHVAR' lschen
        PushVarInstr(Instr);
      end
      else if pos('PUSHCONST', Instr) = 1 then begin
        delete(Instr, 1, 9); //'PUSHCONST' lschen
        PushConstInstr(Instr);
      end
      else raise Exception.Create('ASM: Unknown Instruction Identifier: ' + Instr);
    end;
  end;

  //Erste Anweisung erkennen (Vergleich, Zuweisung, etc) und den Rest an IDentifyInstr bergeben;
  procedure IdentifyLine(line: string);
  var
    mode: string;
    i1, index: integer;
  begin
    repeat
    mode:= GetInstr(line);
    until mode <> '';
    IdentifyInstr(line);
    if pos('MODE', mode) = 1 then begin 
      delete(mode, 1, 4);
      index:= -1;
      for i1:= 0 to high(Comparisons) do begin
        if mode = Comparisons[i1].name then begin
          index:= i1;
          break;
        end;
      end;
      if index >= 0 then begin
        //AddInstr(Sub); //steht bereits am Ende, und wird sowieso hinzugefgt. 
        //--Art des Vergleiches festlegen!!!------------------------------------
      end;
      //----Andere Anweisungen compilieren!!!------------------------------------------------------------------------------------------------------------------------
    end;
  end;

begin
  IdentifyLine(code);
end;

procedure TParser.compile;
var
  i1, mode: integer;
  ASMInstr: string;
  BinInstr: TRTProcArr;

  left, right, expr: string;
  count, stacklen: integer;

  (*
    Gltige Werte:
     0: '='
     1: '>='
     2: '>'
     3: ':='
    -1: '<='
    -2: '<'
    -3: '<>'
  *)
  function GetMode(X: string): integer;
  var
    p: integer;
    mode: string;
  begin
    result:= high(Integer); //mit irgendeinem Schrott fllen
    p:= pos('=', X);
    if p = 0 then p:= pos('<', X);
    if p = 0 then p:= pos('>', X);
    mode:= X[p];
    if mode = '=' then begin
      if X[p - 1] = '<' then result:= -1
      else if X[p - 1] = '>' then result:= 1
      else if X[p - 1] = ':' then result:= 3
      else result:= 0;
    end
    else if mode = '<' then begin
      if X[p + 1] = '>' then result:= -3
      else result:= -2;
    end
    else if mode = '>' then begin
      result:= 2;
    end;
  end;

  procedure AddInstr(X: TRTProcArr);
  var
    xlen, bclen: integer;
    i1: integer;
  begin
    xlen:= length(X);
    bclen:= length(BinCode);
    SetLength(BinCode, bclen + xlen);
    for i1:= 0 to xlen - 1 do begin
      BinCode[bclen + i1]:= X[i1];
    end;
  end;

  procedure SplitAssignment(var instr: string; out vars: string);
  var p: integer;
  begin
    p:= pos(':=', instr);
    if p = 0 then raise Exception.Create(instr + ' contains no assignment');
    vars:= copy(instr, 1, p - 1); //':' soll nicht mehr dabei sein
    delete(instr, 1, p + 1); //':=' soll auch hier nicht dabei sein
  end;

  function GetRightmostExpr(var a: string): string;
  var j1: integer;
  begin
    for j1:= length(a) downto 1 do begin //Leerzeichen am Ende lschen
      if a[j1] <> ' ' then break;
      a:= copy(a, 1, j1 - 1); //dieses Zeichen lschen
    end;
    for j1:= length(a) downto 1 do begin
      if a[j1] = ' ' then break; //Ausdruck hier beendet
      result:= a[j1] + result;   //schreibt von rechtr nach links
      a:= copy(a, 1, j1 - 1); //dieses Zeichen lschen, es ist im nchsten Ausdruck nicht mehr enthalten
    end;
  end;

  function GetStackLength: integer;
  var j1, j2: integer;
  begin
    result:= 2; //weil zwei Variablen von Anfang an da sind
    j1:= 0;
    while j1 <= high(BinCode) do begin
      if @BinCode[j1] = @PushVar then begin
        inc(j1, 2); //die nchste Anweisung ist Schrott
        inc(result);
        continue;
      end;
      if @BinCode[j1] = @PushConst then begin
        inc(j1, 3); //die nchsten 2 Anweisungen sind Schrott
        inc(result);
        continue;
      end;
      for j2:= 0 to high(Operators) do begin
        if (@BinCode[j1] = @Operators[j2].code) then begin
          result:= result + Operators[j2].stacklengthchange;
          break; //subtraktion und = teilen sich eine Funktion ==> wrde doppelt gezhlt werden
        end;
      end;
      inc(j1);
    end;
  end;

begin
  mode:= 0;
  for i1:= 0 to high(code) do begin
    mode:= GetMode(code[i1]);

    if (mode >= -3) and (mode <= 2) then begin
      ParseComp(code[i1], ASMInstr);
      Assemble(ASMInstr, Operators, Comparisons, BinInstr);
      AddInstr(BinInstr);
    end;

    if mode = 3 then begin
      right:= code[i1];
      SplitAssignment(right, left);
      ParseInstr(right, ASMInstr);
      Assemble('MODE:=' + ASMInstr, Operators, Comparisons, BinInstr);
      AddInstr(BinInstr);
      count:= 1; //Mit High beginnen, nicht mit Length!
      stacklen:= GetStackLength;
      //IgnoreSign(left, ',');  //nur eine Anweisung pro Zeile, daher nicht ntig
      while length(left) > 0 do begin
        expr:= ''; //ntig weil sonst nur die neue Variable angehngt wird ?!?
        expr:= GetRightmostExpr(left);
        if expr[1] in Alphabet then begin
          VarName(expr, stacklen - count);
        end;
        inc(count);
      end;
    end;
  end;
  CompMode:= Mode; //Letzte Anweisung enthlt den Vergleich
end;

procedure TParser.execute;
begin
  IP:= 0; //ggf ndern!!!
  repeat
    BinCode[IP](self);
  until IP > high(BinCode);
  index:= high(stack);
end;

procedure TParser.Draw;
var
  index, xci, yci: integer;
  xmin, xmax, ymin, ymax, xcoord, ycoord: double;

  function min(a, b: Integer): Integer;
  begin
    if a < b then result:= a
    else result:= b;
  end;

  procedure DrawEq;
  var
    i1, i2: integer;
    b: boolean;
    val: double;
  begin
    for i1:= 0 to Width - 1 do
      for i2:= 0 to Height - 1 do begin
        b:= false;
        val:= stack[index][i1, i2];
        if i1 > 0 then begin
          if sign(val) <> sign(stack[index][i1 - 1, i2]) then b:= true;
        end;
        if i1 < Width - 1 then begin
          if sign(val) <> sign(stack[index][i1 + 1, i2]) then b:= true;
        end;
        if i2 > 0 then begin
          if sign(val) <> sign(stack[index][i1, i2 - 1]) then b:= true;
        end;
        if i2 < Height - 1 then begin
          if sign(val) <> sign(stack[index][i1, i2 + 1]) then b:= true;
        end;
        if not IsNaN(val) then if IsZero(val, 1.0e-200) then b:= true;
        if b then Canvas.Pixels[i1, Height - 1 - i2]:= clBlue;
      end;
  end;

  procedure DrawGE;
  var i1, i2: integer;
  begin
    for i1:= 0 to Width - 1 do
      for i2:= 0 to Height - 1 do begin
        if stack[index][i1, i2] >= 0 then Canvas.Pixels[i1, Height - 1 - i2]:= clBlue;
      end;
  end;

  procedure DrawGT;
  var i1, i2: integer;
  begin
    for i1:= 0 to Width - 1 do
      for i2:= 0 to Height - 1 do begin
        if stack[index][i1, i2] > 0 then Canvas.Pixels[i1, Height - 1 - i2]:= clBlue;
      end;
  end;

  procedure DrawLE;
  var i1, i2: integer;
  begin
    for i1:= 0 to Width - 1 do
      for i2:= 0 to Height - 1 do begin
        if stack[index][i1, i2] <= 0 then Canvas.Pixels[i1, Height - 1 - i2]:= clBlue;
      end;
  end;

  procedure DrawLT;
  var i1, i2: integer;
  begin
    for i1:= 0 to Width - 1 do
      for i2:= 0 to Height - 1 do begin
        if stack[index][i1, i2] < 0 then Canvas.Pixels[i1, Height - 1 - i2]:= clBlue;
      end;
  end;
begin
  execute(index);
  xmin:= stack[0][0, 0];
  xMax:= stack[0][Width - 1, 0];
  ymin:= stack[1][0, 0];
  ymax:= stack[1][0, Height - 1];
  xcoord:= -xmin / (xmax - xmin);
  ycoord:= -ymin / (ymax - ymin);
  xci:= round(xcoord * Width);
  yci:= round(ycoord * Height);
  Canvas.Pen.Color:= clBlack;
  Canvas.MoveTo(xci, 0);
  Canvas.LineTo(xci, Height);
  Canvas.MoveTo(0, yci);
  Canvas.LineTo(Width, yci);
  if CompMode = 0 then DrawEq
  else if CompMode = 1 then DrawGE
  else if CompMode = 2 then DrawGT
  else if CompMode = -1 then DrawLE
  else if CompMode = -2 then DrawLT;
end;

constructor TParser.Create;
var i1, i2: integer;
begin
  inherited Create;
  Width:= aWidth;
  Height:= aHeight;
  SetLength(Stack, 2);
  SetLength(Stack[0], Width, Height);
  SetLength(Stack[1], Width, Height);
  for i1:= 0 to Width - 1 do
    for i2:= 0 to Height - 1 do begin
      Stack[0][i1, i2]:= XLo + (i1 / (Width - 1)) * (XHi - XLo);
      Stack[1][i1, i2]:= YLo + (i2 / (Height - 1)) * (YHi - YLo);
    end;
  SetLength(Variables, 2);
  Variables[0]:= Variable('x', 0);
  Variables[1]:= Variable('y', 1);

  SetLength(Operators, 16);
  Operators[0]:= OperatorMake('+', Add, -1);
  Operators[1]:= OperatorMake('-', Sub, -1);
  Operators[2]:= OperatorMake('=', Sub, -1); //Vergleich durch Subtraktion
  Operators[3]:= OperatorMake('*', Mul, -1);
  Operators[4]:= OperatorMake('/', Divide, -1);
  Operators[5]:= OperatorMake('^', Pow, -1);
  Operators[6]:= OperatorMake('log', Log, -1);
  Operators[7]:= OperatorMake('sin', Sin, 0);
  Operators[8]:= OperatorMake('cos', Cos, 0);
  Operators[9]:= OperatorMake('ln', Ln, 0);
  Operators[10]:= OperatorMake('sgn', Sgn, 0);
  Operators[11]:= OperatorMake('abs', Abs, 0);
  Operators[12]:= OperatorMake('trunc', Trunc, 0);
  Operators[13]:= OperatorMake('arcsin', asin, 0);
  Operators[14]:= OperatorMake('arccos', acos, 0);
  Operators[15]:= OperatorMake('mod', Modulo, -1);

  SetLength(Comparisons, 6);
  Comparisons[0]:= Comparison('=', 0);
  Comparisons[1]:= Comparison('>=', 1);
  Comparisons[2]:= Comparison('>', 2);
  Comparisons[3]:= Comparison('<=', -1);
  Comparisons[4]:= Comparison('<', -2);
  Comparisons[5]:= Comparison('<>', -3);
end;

procedure Add;
var i1, i2, hs: integer;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      X.stack[hs - 1][i1, i2]:= X.stack[hs - 1][i1, i2] + X.stack[hs][i1, i2];
    end;
  SetLength(X.stack, hs);
end;

procedure Sub;
var
  i1, i2, hs: integer;
  lval, rval: double;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      lval:= X.stack[hs - 1][i1, i2];
      rval:= X.stack[hs][i1, i2];
      if (IsInfinite(lval) or IsNaN(lval)) and (IsInfinite(rval) or IsNaN(rval)) then X.stack[hs - 1][i1, i2]:= NaN
      else X.stack[hs - 1][i1, i2]:= lval - rval;
    end;
  SetLength(X.stack, hs);
end;

procedure Mul;
var i1, i2, hs: integer;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      X.stack[hs - 1][i1, i2]:= X.stack[hs - 1][i1, i2] * X.stack[hs][i1, i2];
    end;
  SetLength(X.stack, hs);
end;

procedure Divide;
var
  i1, i2, hs: integer;
  divisor: double;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      divisor:= X.stack[hs][i1, i2];
      if IsInfinite(divisor) then X.stack[hs - 1][i1, i2]:= 0
      else if IsNaN(divisor) then X.stack[hs - 1][i1, i2]:= NaN
           else if divisor <> 0 then X.stack[hs - 1][i1, i2]:= X.stack[hs - 1][i1, i2] / divisor
                else X.stack[hs - 1][i1, i2]:= infinity;
    end;
  SetLength(X.stack, hs);
end;

procedure Modulo;
var
  i1, i2, hs: integer;
  divisor: double;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      divisor:= X.stack[hs][i1, i2];
      if not IsInfinite(divisor) then //bei Infinity ist der Rest gleich dem Wert, der schon dasteht
           if IsNaN(divisor) then X.stack[hs - 1][i1, i2]:= NaN
           else if divisor <> 0 then X.stack[hs - 1][i1, i2]:= X.stack[hs - 1][i1, i2] - floor(X.stack[hs - 1][i1, i2] / divisor) * divisor
                else X.stack[hs - 1][i1, i2]:= 0;
    end;
  SetLength(X.stack, hs);
end;

procedure Pow;
var
  i1, i2, hs: integer;
  lval, rval: double;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      lval:= X.stack[hs - 1][i1, i2];
      rval:= X.stack[hs][i1, i2];
      if IsInfinite(lval) or IsInfinite(rval) or IsNaN(lval) or IsNaN(rval) then X.stack[hs - 1][i1, i2]:= NaN
      else if (lval <> 0) or (rval > 0) then if (lval >= 0) or (rval = round(rval)) then X.stack[hs - 1][i1, i2]:= power(lval, rval)
                                             else X.stack[hs - 1][i1, i2]:= NaN //Wurzel aus negativer Zahl (dritte geht auch nicht!!!)-----------------------------
           else X.stack[hs - 1][i1, i2]:= NaN;
    end;
  SetLength(X.stack, hs);
end;

procedure Log;
var
  i1, i2, hs: integer;
  b, val: double;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      b:= X.stack[hs - 1][i1, i2];
      val:= X.stack[hs][i1, i2];
      if IsInfinite(val) or IsInfinite(b) or IsNaN(val) or IsNaN(b) then X.stack[hs - 1][i1, i2]:= NaN
      else if (b > 0) and (val > 0) then if b <> 1 then X.stack[hs - 1][i1, i2]:= LogN(b, val)
                                         else X.stack[hs - 1][i1, i2]:= NaN
           else X.stack[hs - 1][i1, i2]:= -infinity;
    end;
  SetLength(X.stack, hs);
end;

procedure Sin;
var
  i1, i2, hs: integer;
  val: double;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      val:= X.stack[hs][i1, i2];
      if IsInfinite(val) or IsNan(val) then X.stack[hs][i1, i2]:= NaN
      else X.stack[hs][i1, i2]:= System.Sin(X.stack[hs][i1, i2]);
    end;
end;

procedure Cos;
var
  i1, i2, hs: integer;
  val: double;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      val:= X.stack[hs][i1, i2];
      if IsInfinite(val) or IsNan(val) then X.stack[hs][i1, i2]:= NaN
      else X.stack[hs][i1, i2]:= System.Cos(X.stack[hs][i1, i2]);
    end;
end;

procedure Ln;
var
  i1, i2, hs: integer;
  val: double;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      val:= X.stack[hs][i1, i2];
      if val > 0 then if not IsInfinite(val) then if not IsNaN(val) then X.stack[hs][i1, i2]:= System.Ln(val)
                                                  else X.stack[hs][i1, i2]:= NaN
                      else X.stack[hs][i1, i2]:= infinity
      else X.stack[hs][i1, i2]:= -infinity;
    end;
end;

procedure Sgn;
var i1, i2, hs: integer;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      X.stack[hs][i1, i2]:= sign(X.stack[hs][i1, i2]);
    end;
end;

procedure Abs;
var i1, i2, hs: integer;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      X.stack[hs][i1, i2]:= System.abs(X.stack[hs][i1, i2]);
    end;
end;

procedure Trunc;
var
  i1, i2, hs: integer;
  val: double;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      val:= X.stack[hs][i1, i2];
      if IsInfinite(val) then X.stack[hs][i1, i2]:= infinity
      else if IsNaN(val) then X.stack[hs][i1, i2]:= NaN
      else X.stack[hs][i1, i2]:= System.trunc(val);
    end;
end;

procedure asin;
var
  i1, i2, hs: integer;
  val: double;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      val:= X.stack[hs][i1, i2];
      if IsInfinite(val) then X.stack[hs][i1, i2]:= infinity
      else if IsNaN(val) then X.stack[hs][i1, i2]:= NaN
      else if (val >= -1) and (val <= 1) then X.stack[hs][i1, i2]:= arcsin(val)
      else X.stack[hs][i1, i2]:= sign(val) * infinity;
    end;
end;

procedure acos;
var
  i1, i2, hs: integer;
  val: double;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      val:= X.stack[hs][i1, i2];
      if IsInfinite(val) then X.stack[hs][i1, i2]:= infinity
      else if IsNaN(val) then X.stack[hs][i1, i2]:= NaN
      else if (val >= -1) and (val <= 1) then X.stack[hs][i1, i2]:= arccos(val)
      else X.stack[hs][i1, i2]:= - sign(val) * infinity;
    end;
end;

procedure Chs;
var i1, i2, hs: integer;
begin
  inc(X.IP);
  hs:= high(X.stack);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      X.stack[hs][i1, i2]:= -(X.stack[hs][i1, i2]);
    end;
end;

procedure PushVar;
var i1, i2, hs, vi: integer;
begin
  SetLength(X.stack, length(X.stack) + 1);
  hs:= high(X.stack);
  SetLength(X.stack[hs], X.Width, X.Height);
  vi:= integer(@X.BinCode[X.IP + 1]); //Parameter (index) wird in der nchsten Anweisung gespeichert, @ gibt den eigentlichen gespeicherten Wert
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      X.stack[hs][i1, i2]:= X.stack[vi][i1, i2]; //Werte kopieren;
    end;
  inc(X.IP, 2); //nchste Anweisung berspringen, dort wre nur Datenmll
end;

procedure PushConst;
var
  i1, i2, hs: integer;
  val: double;
  valBits: array[0..1] of integer absolute val; //Parameter wird in den nchsten zwei Anweisungen gespeichert
begin
  SetLength(X.stack, length(X.stack) + 1);
  hs:= high(X.stack);
  SetLength(X.stack[hs], X.Width, X.Height);
  valBits[0]:= integer(@X.BinCode[X.IP + 1]);
  valBits[1]:= integer(@X.BinCode[X.IP + 2]);
  for i1:= 0 to X.Width - 1 do
    for i2:= 0 to X.Height - 1 do begin
      X.stack[hs][i1, i2]:= val;
    end;
  inc(X.IP, 3); // Nchste 2 Anweisungen berspringen
end;

function Variable;
begin
  result.name:= aName;
  result.index:= aIndex;
end;

function OperatorMake;
begin
  result.name:= aName;
  result.code:= aCode;  
  result.stacklengthchange:= aStackLengthChange;
end;

function Comparison;
begin
  result.name:= aName;
  result.CompMode:= aCompMode;
end;

end.
